#ifndef NVIFILTER_H
#define NVIFILTER_H

#ifndef M_PI
#define M_PI	3.14159265358979323846264338327950288f
#endif

// Filter2D is the abstract data type for a 2D filter.
// width and height are the total width and height of the filter
// The filters are NOT expected to have their results normalized.

class NVIFilter2D {
public:
    NVIFilter2D(float width, float height) : w(width), h(height) {}
    virtual ~NVIFilter2D() {}

    // Set and get the width and height of the filter
    float width (void) const { return w; }
    void width (float newwidth) { w = newwidth; }
    float height (void) const { return h; }
    void height (float newheight) { h = newheight; }
    
    // Evalutate the filter at an x and y position (relative to filter center)
    virtual float operator() (float x, float y) const = 0;

    // Return the name of the filter, e.g., "box", "gaussian"
    virtual const char *name() const = 0;

protected:
    float w, h;
};


class NVIFilterBox2D : public NVIFilter2D {
public:
    NVIFilterBox2D (float width, float height) : NVIFilter2D(width, height) { }
    ~NVIFilterBox2D () { }
    float operator() (float x, float y) const {
        if (fabsf(x) <= w*0.5f && fabsf(y) <= h*0.5f)
            return 1.0f;
        else
            return 0.0f;
    }
    const char *name () const { return "box"; }
};

class NVIFilterTriangle2D : public NVIFilter2D {
public:
    NVIFilterTriangle2D (float width, float height) : NVIFilter2D(width, height) { }
    ~NVIFilterTriangle2D () { }
    float operator() (float x, float y) const {
        return tri1d (x / (w*0.5f)) * tri1d (y / (h*0.5f));
    }
    const char *name () const { return "triangle"; }
private:
    static float tri1d (float x) {
        x = fabsf(x);
        return (x < 1.0f) ? (1.0f - x) : 0.0f;
    }
};

class NVIFilterGaussian2D : public NVIFilter2D {
public:
    NVIFilterGaussian2D (float width, float height) : NVIFilter2D(width, height) { }
    ~NVIFilterGaussian2D () { }
    float operator() (float x, float y) const {
        x = 2.0f * fabsf(x) / w;
        y = 2.0f * fabsf(y) / h;
//        return (x < 1.0 && y < 1.0) ? expf (-2.0f * (x*x+y*y)) : 0.0f;
        return expf (-2.0f * (x*x+y*y));
    }
    const char *name () const { return "gaussian"; }
};

static float smoothstep(float a, float b, float x)
{
    if (x < a)
        return 0.0f;
    if (x > b)
        return 1.0f;
    x = (x - a) / (b - a); // normalize to [0, 1]
    return x*x*(3.0f-2.0f*x);
}

// antialiased disc
class NVIFilterDisc2D : public NVIFilter2D {
public:
    NVIFilterDisc2D (float width, float height, float inner_r = 0.8f) : NVIFilter2D(width, height), _inner_r(inner_r) { }
    ~NVIFilterDisc2D () { }
    float operator() (float x, float y) const {
        x = x / (w*0.5f);
        y = y / (h*0.5f);
        float d = sqrtf(x*x + y*y);
        return 1.0f - smoothstep(_inner_r, 1.0, d);
    }
    const char *name () const { return "disc"; }

private:
    float _inner_r;
};

static float my_atan2(float y, float x) {
    float r = atan2f(y, x);
    if (y < 0.0f)
        return 2.0f*M_PI + r;
    else
        return r;
}

// antialiased regular polygon
class NVIFilterPolygon2D :  public NVIFilter2D {
public:
    NVIFilterPolygon2D (float width, float height, int sides=5, float smooth=0.1f, float rotate=0.0f) : NVIFilter2D(width, height), _sides(sides), _smooth(smooth), _rotate(rotate) { }
    ~NVIFilterPolygon2D () { }

    // point in regular polygon test
    float operator() (float x, float y) const {
        x = x / (w*0.5f);
        y = y / (h*0.5f);
        float segment_ang = 2.0f*M_PI / (float) _sides;
        // rotate point around center
        float rx = cos(_rotate)*x + sin(_rotate)*y;
        float ry = -sin(_rotate)*x + cos(_rotate)*y;
        // figure out which segment we're in
        float ang = my_atan2(ry, rx);
        int segment = (int) (ang / segment_ang);
        // find endpoints of this side
        float a = (segment_ang * segment);
        float a2 = a + segment_ang;
        float x0 = cosf(a); float y0 = sinf(a);
        float x1 = cosf(a2); float y1 = sinf(a2);
        // create perpendicular vector
        float dy = x1 - x0; float dx = y0 - y1;
        float l = 1.0f / sqrtf(dx*dx + dy*dy);
        dx *= l; dy *= l;
        // calculate distance from edge
        float dot = dx*(rx-x0) + dy*(ry-y0);
        //  return dot > 0.0;
        return smoothstep(0.0f, _smooth, dot);
    }
    const char *name () const { return "polygon"; }

    void set_sides(int sides) { _sides = sides; }
    void set_smooth(float smooth) { _smooth = smooth; }
    void set_rotate(float rotate) { _rotate = rotate; }

private:
    int _sides;
    float _smooth, _rotate;
};

class NVIFilterStar2D :  public NVIFilter2D {
public:
    NVIFilterStar2D (float width, float height, int sides=5, float base_width=0.3f, float tip_width=0.0f, float rotate=0.0f, float smooth=0.5f) : NVIFilter2D(width, height), _sides(sides), _base_width(base_width), _tip_width(tip_width), _rotate(rotate), _smooth(smooth) { }
    ~NVIFilterStar2D () { }

    float operator() (float x, float y) const {
        x = x / (w*0.5f);
        y = y / (h*0.5f);
        float segment_ang = 2.0f*M_PI / (float) _sides;
        // rotate point around center
        float rx = cos(_rotate)*x + sin(_rotate)*y;
        float ry = -sin(_rotate)*x + cos(_rotate)*y;
        // figure out which point of the star we're closest too
        float ang = my_atan2(ry, rx);
        int segment = (int) (ang / segment_ang);
        // calculate point vector
        float a = (segment_ang * segment) + segment_ang*0.5f;
        float dy = cosf(a); float dx = -sinf(a);
        // calculate distance from point edge
        float d = sqrtf(rx*rx + ry*ry);
        if (d > 1.0f) d = 1.0f;
        float dot = dx*rx + dy*ry;
        float w = _base_width+(_tip_width - _base_width)*d;
        return (1.0f - smoothstep(w*_smooth, w, fabsf(dot))) * (1.0f - d*d);
    }
    const char *name () const { return "star"; }

    void set_sides(int sides) { _sides = sides; }
    void set_tip_width(float tip_width) { _tip_width = tip_width; }
    void set_base_width(float base_width) { _base_width = base_width; }
    void set_rotate(float rotate) { _rotate = rotate; }
    void set_smooth(float smooth) { _smooth = smooth; }

private:
    int _sides;
    float _base_width, _tip_width;
    float _rotate, _smooth;
};

#endif
