/*
    Image processing - convolutions

    This example shows various methods of implementing convolution filters in OpenGL using Cg.
    It supports both 1D separable filters (such as Gaussian blur) and 2D filters (such as disc blurs).

    The code implements two different methods for implementing convolution in a fragment program. The first
    inlines the kernel weights and offsets into the program - this is the most efficient, but means that
    the program must be recompiled when the kernel changes. The second method stores the kernel weights in
    a texture. This is more expensive due the additional texture lookup, but makes it easy to update the
    kernel interactively.

    The code also shows how to take advantage of bilinear texture filtering so that only half the number of
    texture samples are required for separable filters. This is done by offseting the position of the samples
    and modifying the weights accordingly. This method is described in more detail in "GPU Gems 2", chapter 20.
    Note that only GeForce 6 series cards and later support bilinear filtering for floating point textures.
    
    Finally, this example shows how to implement steerable filters, where the width of the filter can vary
    from pixel to pixel. This is implemented using a texture atlas that stores an array of kernels of
    different widths. It can be used to produce lens-blur effects, where the amount of blur varies based on
    depth from the camera.

    sgreen 5/2005
*/

#if defined(WIN32)
#  include <windows.h>
#endif
#include "load_exr.h"

#define GLH_EXT_SINGLE_FILE
#include <glh/glh_extensions.h>
#include <glh/glh_glut.h>

#include "Cg/cg.h"
#include "Cg/cggl.h"

#include "shared/RenderTexture.h"
#include "shared/quitapp.h"
#include "shared/nv_png.h"
#include "shared/data_path.h"
#include <paramgl/paramgl.h>

#include "convolve.h"

using namespace glh;

char *image_filename = "../../../../media/textures/hdr/bottles2.exr";

//int win_w = 1024, win_h = 768;
int win_w = 512, win_h = 512;

float zoom = 1.0;
float tx = 0.0, ty = 0.0;
int old_x, old_y;
int button_state;

// render targets
RenderTexture *buffer[3];
RenderTexture *visible_buffer = 0;

// Cg programs
CGcontext cg_context = 0;
CGprofile fragment_profile;

CGprogram convolve_fprog = 0, convolve_fprog2 = 0;
CGprogram convolve_tex_fprog = 0, convolve_tex_fprog2 = 0;
CGprogram convolve_steerable_fprog = 0;
CGprogram display_fprog, display_alpha_fprog, passthru_fprog, texture_fprog;
CGparameter exposure_param;
CGprogram unsharp_fprog;
CGparameter unsharp_percent_param, unsharp_threshold_param, unsharp_max_threshold_param;
CGprogram quantize_fprog;
CGparameter quantize_levels_param;
CGprogram median_fprog;

// parameters
float exposure = 1.0;
int kernel_radius = 10;
int atlas_nkernels = 20;
float filter_radius = 10.0;
int polygon_sides = 5;
float polygon_rotate = 0;
float depth_scale = 1.0;
float depth_bias = 0.0;
float unsharp_percent = 1.0;
float unsharp_threshold = 0.0, unsharp_max_threshold = 10.0;
float quantize_levels = 10.0;

// toggles
bool separable = true;
bool use_filtering = false;
bool show_kernel = false;
bool use_kernel_tex = true;
bool do_convolve = true;
bool display_sliders = false;
bool do_unsharp_mask = false;
bool do_quantize = false;
bool do_median = false;
bool display_alpha = false;
bool do_lens_blur = false;
bool use_looping = true;

ParamListGL *params;

GLuint image_tex = 0, kernel_tex = 0, kernel_atlas_tex = 0;
Image2D *kernel_img = 0;
int image_w, image_h;
bool fp16_filter = false;
GLenum buffer_target = GL_TEXTURE_RECTANGLE_NV;

NVIFilter2D *filters[] = {
    new NVIFilterBox2D(2.0f, 2.0f),
    new NVIFilterTriangle2D(2.0f, 2.0f),
    new NVIFilterGaussian2D(2.0f, 2.0f),
    new NVIFilterDisc2D(2.0f, 2.0f),
    new NVIFilterPolygon2D(2.0f, 2.0f),
    new NVIFilterStar2D(2.0f, 2.0f)
};

#define NFILTERS (sizeof(filters) / sizeof(NVIFilter2D *))
#define MAX_FILTER_RADIUS 20

int filter_no = 2;
NVIFilter2D *filter = filters[filter_no];

LARGE_INTEGER frequency;

// get extension pointers
void get_exts()
{
    if(! glh_init_extensions(
                            "GL_ARB_multitexture "
                            "WGL_ARB_pbuffer "
                            "WGL_ARB_pixel_format "
                            "WGL_ARB_render_texture "
                            "GL_VERSION_1_2 "
                            "GL_NV_vertex_program "
                            "GL_NV_fragment_program "
                            "GL_NV_float_buffer "
                            "GL_NV_half_float "
//                            "GL_ARB_color_buffer_float "
//                            "GL_ARB_half_float_pixel "
                            ))
    {
        fprintf(stderr, "Error - required extensions were not supported: %s", glh_get_unsupported_extensions());
        quitapp(-1);
    }
    printf("Extensions: %s\n", glGetString(GL_EXTENSIONS));

    fp16_filter = glutExtensionSupported("GL_ARB_texture_float");
    if (fp16_filter) {
        printf("Using fp16 filtering\n");
        use_filtering = true;
    }
}

RenderTexture *CreateBuffer(int w, int h, char *mode, GLenum target)
{
    RenderTexture *buffer = new RenderTexture(mode, w, h, target);
    buffer->Activate();

    glDisable(GL_DITHER);
    glDisable(GL_DEPTH_TEST);
    glClear(GL_COLOR_BUFFER_BIT);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0.0, w, 0.0, h, -1.0, 1.0);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glViewport(0, 0, w, h);

    buffer->Deactivate();
    return buffer;
}

// write known values into buffer for debugging
void test_buffer(RenderTexture *buffer)
{
    buffer->Activate();
    glClear(GL_COLOR_BUFFER_BIT);
    cgGLBindProgram(passthru_fprog);
    cgGLEnableProfile(fragment_profile);
    glColor3f(1, 1, 1);
    glRectf(8, 8, 9, 9);
    cgGLDisableProfile(fragment_profile);
    buffer->Deactivate();
}

void init_opengl()
{
    glDisable(GL_DEPTH_TEST);
    glClearColor(0.0, 0.0, 0.0, 1.0);
}

void create_buffers(int w, int h)
{
    char *format;
    if (fp16_filter) {
        format = "ati_float=16 rgba texture2D";
//        format = "rgba texture2D";
        buffer_target = GL_TEXTURE_2D;
    } else {
//        format = "float=32 rgba textureRECT";
        format = "float=16 rgba textureRECT";
//        format = "rgba textureRECT";
        buffer_target = GL_TEXTURE_RECTANGLE_NV;
    }
    buffer[0] = CreateBuffer(w, h, format, buffer_target);
    buffer[1] = CreateBuffer(w, h, format, buffer_target);
    buffer[2] = CreateBuffer(w, h, format, buffer_target);
}

void cgErrorCallback()
{
    CGerror lastError = cgGetError();
    if(lastError) {
        printf("Cg error: %s\n\n", cgGetErrorString(lastError));
        printf("%s\n", cgGetLastListing(cg_context));
        cgDestroyContext(cg_context);
        exit(-1);
    }
}

void init_cg()
{
    cgSetErrorCallback(cgErrorCallback);
    cg_context = cgCreateContext();

    fragment_profile = cgGLGetLatestProfile(CG_GL_FRAGMENT);
    cgGLSetOptimalOptions(fragment_profile);

	// initialize Cg compiler arguments
    #define MAX_CG_ARGS 5
    const char *args[MAX_CG_ARGS];
	for(int i=0; i<MAX_CG_ARGS; i++) args[i] = 0;
    int arg = 0;
    args[arg++] = strdup("-DCOMPONENTS=4");
    args[arg++] = strdup("-DUSE_HALF");
    if (fp16_filter) {
        args[arg++] = strdup("-Dh4texRECT=tex2D");
        args[arg++] = strdup("-DsamplerRECT=sampler2D");
    }
    args[arg++] = 0;

    data_path media;
    media.path.push_back(".");
    media.path.push_back("../../../MEDIA/programs/imgproc_convolve");
    media.path.push_back("../../../../MEDIA/programs/imgproc_convolve");

	// load programs
    display_fprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, media.get_file("display_hdr.cg").c_str(),
                                            fragment_profile, "display_hdr", args);
    cgGLLoadProgram(display_fprog);
    exposure_param = cgGetNamedParameter(display_fprog, "exposure");

    display_alpha_fprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, media.get_file("display_hdr.cg").c_str(),
                                            fragment_profile, "display_alpha", args);
    cgGLLoadProgram(display_alpha_fprog);

    passthru_fprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, media.get_file("passthru.cg").c_str(),
                                            fragment_profile, "main", NULL);
    cgGLLoadProgram(passthru_fprog);

    texture_fprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, media.get_file("texture.cg").c_str(),
                                            fragment_profile, "texRECT", NULL);
    cgGLLoadProgram(texture_fprog);

    unsharp_fprog = cgCreateProgramFromFile(cg_context, CG_SOURCE,
                                            media.get_file("unsharp_mask.cg").c_str(),
                                            fragment_profile, "unsharp_mask", args);
    cgGLLoadProgram(unsharp_fprog);
    unsharp_percent_param = cgGetNamedParameter(unsharp_fprog, "percent");
    unsharp_threshold_param = cgGetNamedParameter(unsharp_fprog, "threshold");
    unsharp_max_threshold_param = cgGetNamedParameter(unsharp_fprog, "max_threshold");

    quantize_fprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, media.get_file("quantize.cg").c_str(),
                                            fragment_profile, "quantize", args);
    cgGLLoadProgram(quantize_fprog);
    quantize_levels_param = cgGetNamedParameter(quantize_fprog, "levels");

    median_fprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, media.get_file("median.cg").c_str(),
                                           fragment_profile, "median_3x3", args);

    CGparameter texel_size_param = cgGetNamedParameter(median_fprog, "texel_size");
    cgSetParameterVariability(texel_size_param, CG_LITERAL);
    if (fp16_filter) {
        cgSetParameter2f(texel_size_param, 1.0f/image_w, 1.0f/image_h);
    } else {
        cgSetParameter2f(texel_size_param, 1.0f, 1.0f);
    }
    cgGLLoadProgram(median_fprog);

    // free arguments
    for(int i=0; i<MAX_CG_ARGS; i++) {
        if (args[i]) free((void *) args[i]);
    }
}

void DrawQuad(int w, int h, bool rect)
{
    glBegin(GL_QUADS);
    if (rect) {
        glTexCoord2i(0, 0); glVertex2i(0, 0);
        glTexCoord2i(w, 0); glVertex2i(w, 0);
        glTexCoord2i(w, h); glVertex2i(w, h);
        glTexCoord2i(0, h); glVertex2i(0, h);
    } else {
        glTexCoord2i(0, 0); glVertex2i(0, 0);
        glTexCoord2i(1, 0); glVertex2i(w, 0);
        glTexCoord2i(1, 1); glVertex2i(w, h);
        glTexCoord2i(0, 1); glVertex2i(0, h);
    }
    glEnd();
}

// perform convolution from source to destination buffer using given fragment program
void doConvolutionPass(RenderTexture *src_buffer, RenderTexture *dest_buffer, CGprogram prog)
{
    dest_buffer->Activate();

    cgGLBindProgram(prog);
    cgGLEnableProfile(fragment_profile);

    glActiveTextureARB(GL_TEXTURE0_ARB);
    src_buffer->Bind();
    if (use_filtering) {
	    glTexParameteri(buffer_target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	    glTexParameteri(buffer_target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    } else {
	    glTexParameteri(buffer_target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	    glTexParameteri(buffer_target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    }

    glActiveTextureARB(GL_TEXTURE1_ARB);
    glBindTexture(GL_TEXTURE_RECTANGLE_NV, do_lens_blur ? kernel_atlas_tex: kernel_tex);

    DrawQuad(dest_buffer->GetWidth(), dest_buffer->GetHeight(), buffer_target==GL_TEXTURE_RECTANGLE_NV);

    glActiveTextureARB(GL_TEXTURE0_ARB);
    src_buffer->Release();

    cgGLDisableProfile(fragment_profile);

    dest_buffer->Deactivate();
}

// perform separable convolution using two passes
void doSeparableConvolution(RenderTexture *src_buffer, RenderTexture *dest_buffer, RenderTexture *temp_buffer, CGprogram h_fprog, CGprogram v_fprog)
{
    doConvolutionPass(src_buffer, temp_buffer, h_fprog);
    doConvolutionPass(temp_buffer, dest_buffer, v_fprog);
}

void do_unary_op(RenderTexture *src_buffer, RenderTexture *dest_buffer, CGprogram prog)
{
    dest_buffer->Activate();

    cgGLBindProgram(prog);
    cgGLEnableProfile(fragment_profile);

    glActiveTextureARB(GL_TEXTURE0_ARB);
    src_buffer->Bind();

    DrawQuad(dest_buffer->GetWidth(), dest_buffer->GetHeight(), buffer_target==GL_TEXTURE_RECTANGLE_NV);

    glActiveTextureARB(GL_TEXTURE0_ARB);
    src_buffer->Release();

    cgGLDisableProfile(fragment_profile);

    dest_buffer->Deactivate();
}

void do_binary_op(RenderTexture *src_buffer, RenderTexture *src_buffer2, RenderTexture *dest_buffer, CGprogram prog)
{
    dest_buffer->Activate();

    cgGLBindProgram(prog);
    cgGLEnableProfile(fragment_profile);

    glActiveTextureARB(GL_TEXTURE0_ARB);
    src_buffer->Bind();
    glActiveTextureARB(GL_TEXTURE1_ARB);
    src_buffer2->Bind();

    DrawQuad(dest_buffer->GetWidth(), dest_buffer->GetHeight(), buffer_target==GL_TEXTURE_RECTANGLE_NV);

    glActiveTextureARB(GL_TEXTURE0_ARB);
    src_buffer->Release();
    glActiveTextureARB(GL_TEXTURE1_ARB);
    src_buffer2->Release();

    cgGLDisableProfile(fragment_profile);

    dest_buffer->Deactivate();
}

float GetTimeDifference(LARGE_INTEGER *pStart, LARGE_INTEGER *pEnd)
{
    float result;
    result = (float)(pEnd->QuadPart - pStart->QuadPart);
    result /= frequency.QuadPart;
    return result;
}

// perform convolution based on current settings
void do_convolution(RenderTexture *src, RenderTexture *dest, RenderTexture *temp)
{
    LARGE_INTEGER start, end;
    QueryPerformanceCounter(&start);

    if (do_lens_blur) {
        setDepthScaleBias(depth_scale, depth_bias);
        doConvolutionPass(src, dest, convolve_steerable_fprog);

    } else if (separable) {
        if (use_kernel_tex) {
            doSeparableConvolution(src, dest, temp, convolve_tex_fprog, convolve_tex_fprog2);
        } else {
            doSeparableConvolution(src, dest, temp, convolve_fprog, convolve_fprog2);
        }

    } else {
        doConvolutionPass(src, dest, use_kernel_tex ? convolve_tex_fprog : convolve_fprog);
    }

    glFinish();
    QueryPerformanceCounter(&end);
    float time = GetTimeDifference(&start, &end);
    printf("%.3f secs, %.2f Mpixels/sec\n", time, image_w*image_h / (time*1000000.0));
}

// initialize fragment programs for convolution
void init_convolution_prog()
{
    if (convolve_fprog) {
        cgDestroyProgram(convolve_fprog);
        convolve_fprog = 0;
    }
    if (convolve_fprog2) {
        cgDestroyProgram(convolve_fprog2);
        convolve_fprog2 = 0;
    }
    kernel_radius = ceil(filter_radius);

    if (separable) {
        convolve_fprog = createConvolutionProgramSeparable(*filter, kernel_radius, use_filtering, false, fp16_filter, image_w, image_h);
        convolve_fprog2 = createConvolutionProgramSeparable(*filter, kernel_radius, use_filtering, true, fp16_filter, image_w, image_h);
    } else {
        // avoid exceeding maximum program length
        if (kernel_radius > 10) kernel_radius = 10;
        convolve_fprog = createConvolutionProgram2D(*filter, kernel_radius, kernel_radius, fp16_filter, image_w, image_h);
    }
}

// build kernel texture based on current filter settings
void init_kernel_tex()
{
    // create kernel texture
    Image2D *img;
    if (separable) {
        img = evaluateFilter2D(*filter, kernel_radius, 0, true);
    } else {
        img = evaluateFilter2D(*filter, kernel_radius, kernel_radius, true);
    }
    if (kernel_tex) {
        glDeleteTextures(1, &kernel_tex);
    }
    kernel_tex = createTextureFromImage2D(*img, GL_TEXTURE_RECTANGLE_NV, GL_FLOAT_R16_NV);
    delete img;

    // create unnormalized version for display
    if (kernel_img) {
        delete kernel_img;
    }
    if (separable) {
        kernel_img = evaluateFilter2D(*filter, kernel_radius, 0, false);
    } else {
        kernel_img = evaluateFilter2D(*filter, kernel_radius, kernel_radius, false);
    }
}

// initialize fragment program for performing convolution using kernel texture
void init_convolution_prog_tex()
{
    if (convolve_tex_fprog) {
        cgDestroyProgram(convolve_tex_fprog);
        convolve_tex_fprog = 0;
    }
    if (convolve_fprog2) {
        cgDestroyProgram(convolve_fprog2);
        convolve_fprog2 = 0;
    }
    kernel_radius = ceil(filter_radius);

    if (separable) {
        convolve_tex_fprog = createConvolveProgramTexSeparable(kernel_radius*2+1, -kernel_radius, fp16_filter, false, image_w, image_h);
        convolve_tex_fprog2 = createConvolveProgramTexSeparable(kernel_radius*2+1, -kernel_radius, fp16_filter, true, image_w, image_h);
    } else {
        // avoid exceeding maximum program length
        if (((fragment_profile==CG_PROFILE_FP30) || (fragment_profile==CG_PROFILE_FP40 && !use_looping))
            && (kernel_radius > 7)) kernel_radius = 7;
        convolve_tex_fprog = createConvolveProgramTex(kernel_radius*2+1, kernel_radius*2+1, -kernel_radius, -kernel_radius, fp16_filter, image_w, image_h);
    }
}

// initialize kernel atlas texture for steerable filter
void init_kernel_atlas_tex(int nkernels, int max_radius)
{
    // avoid exceeding max fragment program length on NV3x
    if (((fragment_profile==CG_PROFILE_FP30) || (fragment_profile==CG_PROFILE_FP40 && !use_looping))
        && (max_radius > 5)) max_radius = 5;

    // create kernel texture atlas
    Image2D *img= evaluateFilterAtlas2D(*filter, nkernels, max_radius, 1.0, max_radius);
    if (kernel_atlas_tex) {
        glDeleteTextures(1, &kernel_atlas_tex);
    }
    kernel_atlas_tex = createTextureFromImage2D(*img, GL_TEXTURE_RECTANGLE_NV, GL_FLOAT_R16_NV);
    delete img;
}

// initialize fragment program for performing steerable convolution
void init_convolution_prog_steerable(int nkernels, int max_radius)
{
    if (((fragment_profile==CG_PROFILE_FP30) || (fragment_profile==CG_PROFILE_FP40 && !use_looping))
        && (max_radius > 5)) max_radius = 5;

    // create fragment program
    if (convolve_steerable_fprog) {
        cgDestroyProgram(convolve_steerable_fprog);
        convolve_steerable_fprog = 0;
    }
    convolve_steerable_fprog = createConvolveProgramTexSteerable(max_radius*2+1, max_radius*2+1, -max_radius, -max_radius, nkernels, fp16_filter, image_w, image_h);
}

void set_filter_params()
{
    filter->width(filter_radius*2);
    filter->height(filter_radius*2);
    ((NVIFilterPolygon2D *) filters[4])->set_sides(polygon_sides);
    ((NVIFilterStar2D *) filters[5])->set_sides(polygon_sides);
    ((NVIFilterPolygon2D *) filters[4])->set_rotate(polygon_rotate);
    ((NVIFilterStar2D *) filters[5])->set_rotate(polygon_rotate);
}

// initialize convolution fragment programs and textures based on current settings
void init_convolution()
{
    set_filter_params();

    if (do_lens_blur) {
        init_kernel_atlas_tex(atlas_nkernels, kernel_radius);
        init_convolution_prog_steerable(atlas_nkernels, kernel_radius);
    } else if (use_kernel_tex) {
        init_kernel_tex();
        init_convolution_prog_tex();
    } else {
        init_convolution_prog();
    }
    printf("filter radius = %.2f, kernel radius = %d\n", filter_radius, kernel_radius);
    printf("filtering: %d, separable: %d, kernel_texture: %d, looping: %d\n", use_filtering, separable, use_kernel_tex, use_looping);
    do_convolution(buffer[0], buffer[1], buffer[2]);
}

void display_kernel()
{
#if 1
    glRasterPos2i(0, 0);
    glPixelZoom(zoom, zoom);
    glDrawPixels(kernel_img->get_width(), kernel_img->get_height(), GL_LUMINANCE, GL_FLOAT, (const void *) kernel_img->get_pointer());
#else
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glBindTexture(GL_TEXTURE_RECTANGLE_NV, kernel_tex);
    cgGLBindProgram(texture_fprog);
    cgGLEnableProfile(fragment_profile);
    DrawQuad(kernel_img->get_width(), kernel_img->get_width(), true);
    cgGLDisableProfile(fragment_profile);
#endif
}

void display_buffer(RenderTexture *buffer)
{
    glActiveTextureARB(GL_TEXTURE0_ARB);
    buffer->Bind();
	glTexParameteri(buffer_target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(buffer_target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    if (display_alpha) {
        cgGLBindProgram(display_alpha_fprog);
    } else {
        cgGLBindProgram(display_fprog);
        cgSetParameter1f(exposure_param, exposure);
    }
    cgGLEnableProfile(fragment_profile);
    DrawQuad(buffer->GetWidth(), buffer->GetHeight(), buffer_target==GL_TEXTURE_RECTANGLE_NV);
    cgGLDisableProfile(fragment_profile);
    buffer->Release();
}

void do_processing()
{
    // median filter
    if (do_median) {
        do_unary_op(buffer[0], buffer[1], median_fprog);
        return;
    }

    // convolution
    if (do_convolve) {
        set_filter_params();
        if (do_lens_blur) {
            init_kernel_atlas_tex(atlas_nkernels, kernel_radius);
        } else {
            init_kernel_tex();
        }
        do_convolution(buffer[0], buffer[1], buffer[2]);
    }

    // quantize
    if (do_quantize) {
        cgSetParameter1f(quantize_levels_param, quantize_levels);
        do_unary_op(buffer[1], buffer[2], quantize_fprog);
        return;
    }

    // unsharp mask
    if (do_unsharp_mask) {
        cgSetParameter1f(unsharp_percent_param, unsharp_percent);
        cgSetParameter1f(unsharp_threshold_param, unsharp_threshold);
        cgSetParameter1f(unsharp_max_threshold_param, unsharp_max_threshold);
        do_binary_op(buffer[0], buffer[1], buffer[2], unsharp_fprog);
    }

}

void display()
{
    do_processing();

    // display buffer
    glClear(GL_COLOR_BUFFER_BIT);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glScalef(zoom, zoom, 1.0);
    glTranslatef(tx, ty, 0.0);

    if (show_kernel) {
        display_kernel();
    } else {
        display_buffer(visible_buffer);
    }

    if (display_sliders) {
        // sliders
        glBlendFunc(GL_ONE_MINUS_DST_COLOR, GL_ZERO); // invert color
        glEnable(GL_BLEND);
        params->Render(0, 0);
        glDisable(GL_BLEND);
    }

    glutSwapBuffers();
    glutReportErrors();
}

// dump buffer values for debugging
void dump_buffer(RenderTexture *buffer)
{
    int w = visible_buffer->GetWidth();
    buffer->Activate();
    float *data = new float [w];
    glReadPixels(0, 8, w, 1, GL_RED, GL_FLOAT, data);
    for(int i=0; i<w; i++) {
        printf("%d: %f\n", i, data[i]);
    }
    delete [] data;
    buffer->Deactivate();
}

void idle()
{    
//    glutPostRedisplay();
}

void key(unsigned char k, int x, int y)
{
    switch(k) {
        case 27:
            exit(0);
            break;
        case '=':
        case '+':
            zoom *= 2.0;
            break;
        case '-':
            zoom *= 0.5;
            break;

        case '1':
        case '2':
        case '3':
            visible_buffer = buffer[k - '1'];
            break;

        case 'c':
            do_convolution(buffer[0], buffer[1], buffer[2]);
            break;

        case 'r':
            // reset view
            tx = ty = 0.0;
            zoom = 1.0;
            break;

        case 'd':
            dump_buffer(visible_buffer);
            break;

        case 'p':
            {
                // dump program
                CGprogram prog;
                if (do_lens_blur) {
                    prog = convolve_steerable_fprog;
                } else if (use_kernel_tex) {
                    prog = convolve_tex_fprog;
                } else {
                    prog = convolve_fprog;
                }
                const char *obj_code = cgGetProgramString(prog, CG_COMPILED_PROGRAM);
                printf("%s\n", obj_code);
            }
            break;

        case 't':
            use_kernel_tex ^= 1;
            do_convolve = use_kernel_tex;
            init_convolution();
            break;

        case '.':
            filter_radius+=0.5;
//            if (filter_radius > MAX_FILTER_RADIUS) filter_radius = MAX_FILTER_RADIUS;
            init_convolution();
            break;
        case ',':
            filter_radius-=0.5;
//            if (filter_radius < 0) filter_radius = 0; 
            init_convolution();
            break;

        case ']':
            exposure *= 2.0;
            printf("exposure = %f\n", exposure);
            break;
        case '[':
            exposure /= 2.0;
            printf("exposure = %f\n", exposure);
            break;

        case 's':
            separable = !separable;
            init_convolution();
            break;

        case 'f':
            use_filtering = !use_filtering;
            init_convolution();
            break;

        case '\'':
            polygon_sides++;
            init_convolution();
            break;
        case ';':
            if (polygon_sides > 3) polygon_sides--;
            init_convolution();
            break;

        case '"':
            polygon_rotate += 0.1;
            init_convolution();
            break;
        case ':':
            polygon_rotate -= 0.1;
            init_convolution();
            break;

        case 'k':
            show_kernel ^= 1;
            break;

        case 'h':
            display_sliders ^= 1;
            break;

        case 'u':
            do_unsharp_mask ^= 1;
            visible_buffer = do_unsharp_mask ? buffer[2] : buffer[1];
            filter = filters[2]; // gaussian
            filter_radius = 2.0;
            break;

        case 'q':
            do_quantize ^= 1;
            visible_buffer = do_quantize ? buffer[2] : buffer[1];
            break;

        case 'm':
            do_median ^= 1;
            break;

        case 'a':
            display_alpha ^= 1;
            break;

        case 'l':
            do_lens_blur ^= 1;
            init_convolution();
            break;

        case 'o':
            use_looping ^= 1;
            destroy_convolve();
            init_convolve(true, 4, fp16_filter, use_looping);
            init_convolution();
            break;
    }
    
	glutPostRedisplay();
}

void special(int key, int x, int y)
{
    if (display_sliders) {
        params->Special(key, x, y);
        return;
    }

    switch(key) {
    case GLUT_KEY_RIGHT:
        tx += 1.0;
        break;
    case GLUT_KEY_LEFT:
        tx -= 1.0;
        break;
    case GLUT_KEY_UP:
        ty += 1.0;
        break;
    case GLUT_KEY_DOWN:
        ty -= 1.0;
        break;
    }
    glutPostRedisplay();
}


void resize(int w, int h)
{
    glViewport(0, 0, w, h);
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0.0, w, 0.0, h, -1.0, 1.0);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

void mouse(int button, int state, int x, int y)
{
    old_x = x; old_y = y;
    if (state == GLUT_DOWN) {
        button_state |= 1<<button;
    } else if (state == GLUT_UP) {
        button_state = 0;
    }

    if (display_sliders && (button_state & 1)) {
        // call list mouse function
        if (params->Mouse(x, y, button, state)) {
            glutPostRedisplay();
            return;
        }
    }

}

void motion(int x, int y)
{
    if (display_sliders) {
        // call list motion function
        if (params->Motion(x, y)) {
            glutPostRedisplay();
            return; 
        };
    }

    if (button_state == 1) {
        tx += (x - old_x) / zoom;
        ty -= (y - old_y) / zoom;
    } else if (button_state == 2) {
        zoom += (y - old_y) * 0.01;
    } else if (button_state == 3) {
    }

    old_x = x; old_y = y;
    glutPostRedisplay();
}

void main_menu(int i)
{
    key((unsigned char) i, 0, 0);
}

void filter_menu(int i)
{
    filter_no = i;
    filter = filters[filter_no];
    if (i >= 3) {
        separable = false;
    } else {
        separable = true;
    }

    init_convolution();

    glutPostRedisplay();
}

void radius_menu(int i)
{
    filter_radius = i;
    init_convolution();
    glutPostRedisplay();
}

void init_menu()
{    
    int filtermenu = glutCreateMenu(filter_menu);
    for(int i=0; i<NFILTERS; i++) {
        glutAddMenuEntry(filters[i]->name(), i);
    }
 
    int radiusmenu = glutCreateMenu(radius_menu);
    for(int i=0; i<MAX_FILTER_RADIUS; i++) {
        char str[10];
        sprintf(str, "%d", i+1);
        glutAddMenuEntry(str, i+1);
    }

    int togglemenu = glutCreateMenu(main_menu);
    glutAddMenuEntry("Toggle separable filter [s]", 's');
    glutAddMenuEntry("Toggle bilinear filtering [f]", 'f');
    glutAddMenuEntry("Toggle use kernel texture [t]", 't');
    glutAddMenuEntry("Toggle shader looping [o]", 'o');
    glutAddMenuEntry("Toggle show kernel [k]", 'k');
    glutAddMenuEntry("Toggle show alpha [a]", 'a');
    glutAddMenuEntry("Toggle unsharp mask [u]", 'u');
    glutAddMenuEntry("Toggle quantize [q]", 'q');
    glutAddMenuEntry("Toggle lens blur mode [l]", 'l');
    glutAddMenuEntry("Toggle median filter mode [m]", 'm');

    glutCreateMenu(main_menu);
    glutAddSubMenu("Filter", filtermenu);
    glutAddSubMenu("Kernel radius", radiusmenu);
    glutAddSubMenu("Toggles", togglemenu);

    glutAddMenuEntry("Show source buffer [1]", '1');
    glutAddMenuEntry("Show destination buffer [2]", '2');
    glutAddMenuEntry("Show temporary buffer [3]", '3');
    glutAddMenuEntry("Zoom in [=]", '=');
    glutAddMenuEntry("Zoom out [-]", '-');
    glutAddMenuEntry("Increase radius [.]", '.');
    glutAddMenuEntry("Decrease radius [,]", ',');
    glutAddMenuEntry("Double exposure []]", ']');
    glutAddMenuEntry("Half exposure [[]", '['); 
    glutAddMenuEntry("Time convolution [c]", 'c');
    glutAddMenuEntry("Dump fragment program [p]", 'p');
    glutAddMenuEntry("Toggle sliders [h]", 'h');
    glutAddMenuEntry("Reset view [r]", 'r');
    glutAddMenuEntry("Quit [esc]", '\033');
    glutAttachMenu(GLUT_RIGHT_BUTTON);
}

// create sliders
void init_params()
{
  params = new ParamListGL("");

  params->AddParam(new Param<float>("filter radius", filter_radius, 1.0, MAX_FILTER_RADIUS, 0.1, &filter_radius));
  params->AddParam(new Param<float>("exposure", 1.0, 0.01, 16.0, 0.001, &exposure));
  params->AddParam(new Param<int>("polygon sides", polygon_sides, 3, 10, 1, &polygon_sides));
  params->AddParam(new Param<float>("polygon rotate", polygon_rotate, 0.0, 2.0, 0.01, &polygon_rotate));
  params->AddParam(new Param<float>("unsharp threshold", unsharp_threshold, 0.0, 10.0, 0.01, &unsharp_threshold));
  params->AddParam(new Param<float>("unsharp max threshold", unsharp_max_threshold, 0.0, 10.0, 0.01, &unsharp_max_threshold));
  params->AddParam(new Param<float>("unsharp percent", unsharp_percent, 0.0, 1.0, 0.01, &unsharp_percent));
  params->AddParam(new Param<float>("quantize levels", quantize_levels, 0.0, 20.0, 1.0, &quantize_levels));
  params->AddParam(new Param<float>("depth scale", 1.0, 0.0, 10.0, 0.01, &depth_scale));
  params->AddParam(new Param<float>("depth bias", 0.0, -2.0, 2.0, 0.01, &depth_bias));
}

GLuint create_texture(GLenum target, GLenum internal_format, GLsizei w, GLsizei h, GLenum format, GLenum type, void *data, bool filter)
{
    GLuint tex;
    glGenTextures(1, &tex);
    glBindTexture(target, tex);
    if (filter) {
	    glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	    glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    } else {
	    glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	    glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    }
    glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glTexImage2D(target, 0, internal_format, w, h, 0, format, type, data);
    return tex;
}

void *load_png(char *filename, int &width, int &height, GLenum &format, GLenum &type)
{
    array2<vec3ub> img;
    read_png_rgb(filename, img);
    width = img.get_width(); height = img.get_height();
    format = GL_RGB;
    type = GL_UNSIGNED_BYTE;
    return img.get_pointer();
}

void *load_exr(char *filename, int &width, int &height, GLenum &format, GLenum &type)
{
    Imf::Array2D<Imf::Rgba> *pixels;
    pixels = load_OpenEXR(filename, width, height);
    format = GL_RGBA;
    type = GL_HALF_FLOAT_NV;
    return (*pixels)[0];
}

int main(int argc, char **argv)
{
	glutInit(&argc, argv);
	glutInitWindowSize(win_w, win_h);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB);
	glutCreateWindow("Convolution filters");

    get_exts();
	init_opengl();
    init_menu();
    init_params();

    if (argc > 1) {
        image_filename = argv[1];
    }

#if 0
    // debug
	image_w = image_h = 16;
    init_cg();
    create_buffers(image_w, image_h);
    test_buffer(buffer[0]);
    zoom = 32.0;
#else
    // load image
    GLenum format, type;
    void *data;
#if USE_PNG
    data = load_png(image_filename, image_w, image_h, format, type)
#else
    data = load_exr(image_filename, image_w, image_h, format, type);
#endif

    printf("Loaded '%s', %d x %d pixels\n", image_filename, image_w, image_h);
    create_buffers(image_w, image_h);

    init_cg();

    // render image to source buffer
    buffer[0]->Activate();
    cgGLBindProgram(passthru_fprog);
    cgGLEnableProfile(fragment_profile);
    glDrawPixels(image_w, image_h, format, type, data);
    cgGLDisableProfile(fragment_profile);
    buffer[0]->Deactivate();
#endif

    init_convolve(true, 4, fp16_filter, use_looping);
    init_convolution();

    visible_buffer = buffer[1];
    display_sliders = true;
    QueryPerformanceFrequency(&frequency);

	glutDisplayFunc(display);
    glutMouseFunc(mouse);
    glutMotionFunc(motion);
    glutIdleFunc(idle);
    glutKeyboardFunc(key);
    glutSpecialFunc(special);
    glutReshapeFunc(resize);

	glutMainLoop();
	return 0;
}
